@@ -60,6 +60,7 @@ typedef struct {
   long max_concurrent_xfrs;  // dsmgrd.parallelxfrs
   long max_stage_retries;    // dsmgrd.corruptafterfails
   long cmd_timeout_secs;     // dsmgrd.cmdtimeoutsecs
+  bool purge_noop_ds;        // dsmgrd.purgenoopds
   std::string stage_cmd;     // dsmgrd.stagecmd
   af::regex **url_regexs;    // dsmgrd.urlregex[n]
   unsigned int n_url_regexs;
@@ -432,6 +433,7 @@ void process_datasets(af::opQueue &opq, af::dataSetList &dsm,
   const char *ds;
   dsm.fetch_datasets();
   unsigned int count_ds = 0;
+  unsigned int deleted_ds = 0;
 
   while (ds = dsm.next_dataset()) {
 
@@ -642,23 +644,52 @@ void process_datasets(af::opQueue &opq, af::dataSetList &dsm,
     // Save only if needed
     //
 
+    bool nothing_done = true;
+
     if (count_changes > 0) {
 
-      //toggle_suid();
       bool save_ok = dsm.save_dataset();
-      //toggle_suid();
 
       if (save_ok) {
-        af::log::ok(af::log_level_high, "Dataset %s saved: %d entries", ds,
-          count_files);
+        af::log::ok(af::log_level_high,
+          "Dataset %s saved: %d entries considered", ds, count_files);
       }
       else {
         af::log::error(af::log_level_high,
           "Dataset %s not saved: check permissions", ds);
       }
+
+      nothing_done = false;
     }
-    else {
-      af::log::info(af::log_level_low, "Dataset %s not modified", ds);
+    else if ((vars.purge_noop_ds) && (count_files == 0)) {
+
+      // If there's at least one corrupted file, don't delete dataset
+      dsm.free_files();  // free prev resources: we allocate new ones
+      dsm.fetch_files(NULL, "C");
+      if (dsm.next_file() == NULL) {
+        // No corrupted files
+        af::log::info(af::log_level_debug, "Dataset %s is condemned", ds);
+        if (dsm.remove_dataset(ds)) {
+          af::log::ok(af::log_level_low, "Dataset %s deleted", ds);
+          deleted_ds++;
+        }
+        else {
+          af::log::error(af::log_level_high, "Failed to delete dataset %s", ds);
+        }
+        nothing_done = false;
+      }
+      else {
+        af::log::info(af::log_level_debug,
+          "Dataset %s has corrupted files: not deleted", ds);
+      }
+    }
+
+    dsm.free_files();
+
+    if (nothing_done) {
+      af::log::info(af::log_level_low,
+        "Dataset %s not modified: %d entries were considered",
+        ds, count_files);
     }
 
     if (vars.notif) {
@@ -697,8 +728,9 @@ void process_datasets(af::opQueue &opq, af::dataSetList &dsm,
 
   dsm.free_datasets();
 
-  af::log::info(af::log_level_low, "Number of datasets processed: %u",
-    count_ds);
+  af::log::info(af::log_level_low,
+    "Number of datasets processed: %u (deleted: %u)",
+    count_ds, deleted_ds);
 
   //
   // Clean up transfer queue
@@ -759,6 +791,7 @@ void main_loop(af::config &config) {
     AF_INT_MAX);  // 0 == timeout off
   config.bind_callback("dsmgrd.notifyplugin", &config_callback_notify,
     notif_cbk_args);
+  config.bind_bool("dsmgrd.purgenoopds", &vars.purge_noop_ds, false);
 
   // Initializes regular expression objects for URL substitutions and their
   // respective callbacks
@@ -816,6 +849,8 @@ void main_loop(af::config &config) {
 
       // Select the one to consider: stgreq has priority
       if (!dsm_stgreq_path.empty()) {
+        if (strncmp("dir:", dsm_stgreq_path.c_str(), 4) == 0)
+          dsm_stgreq_path.erase(0, 4);
         dsm_new_path = &dsm_stgreq_path;
         dsm_from_stgreq = true;
       }
@@ -826,7 +861,7 @@ void main_loop(af::config &config) {
 
         if (dsm_path.empty()) {
           af::log::error(af::log_level_urgent,
-            "None of xpd.datasetsrc or xpd.stagereqrepo contain a "
+            "None of xpd.datasetsrc or xpd.stagereqrepo contains a "
             "valid dataset repository path!");
           dsm.set_dataset_mgr(NULL);
         }
@@ -837,7 +872,7 @@ void main_loop(af::config &config) {
 
           TDataSetManagerFile *root_dsm = new TDataSetManagerFile(NULL, NULL,
             root_dsm_opts.c_str());
-          dsm.set_dataset_mgr(root_dsm);  // has ownership
+          dsm.set_dataset_mgr(root_dsm, dsm_path.c_str());  // has ownership
           af::log::ok(af::log_level_urgent,
             "New ROOT dataset manager initialized with options: %s",
             root_dsm_opts.c_str());
